"""A python package for Command & control (C2) of Autonomous Random finite set- (RFS-) Based Swarms.

This code is developed by the Laboratory for Autonomy, Guidance, navigation,
control, and Estimation Research (LAGER) at the University of Alabama.
"""
